/***************************************
 *            ViPER                    *
 *  The Video Processing               *
 *         Evaluation Resource         *
 *                                     *
 *  Distributed under the GPL license  *
 *        Terms available at gnu.org.  *
 *                                     *
 *  Copyright University of Maryland,  *
 *                      College Park.  *
 ***************************************/

package edu.umd.cfar.lamp.viper.examples.omegahead;

import java.awt.*;
import java.awt.geom.*;
import java.util.logging.*;

import viper.api.*;
import viper.api.time.*;
import edu.umd.cfar.lamp.viper.geometry.*;
import edu.umd.cfar.lamp.viper.gui.canvas.*;
import edu.umd.cfar.lamp.viper.gui.canvas.datatypes.*;
import edu.umd.cfar.lamp.viper.gui.core.*;
import edu.umd.cs.piccolo.nodes.*;

/**
 * Displays an 'omega' shape on the image.
 * Right now, this is just a circle above a line.
 * 
 */
public class OmegaNode extends AttributablePPathAdapter implements Attributable {
	private static Logger logger = Logger
			.getLogger("edu.umd.cfar.lamp.viper.gui.canvas.datatypes");
	private static final double HANDLE_LENGTH = 20;
	private static Shape EMPTY_SHAPE = new GeneralPath();
	
	
	/** Enumeration for the type of highlight currently applied */
	public static interface HIGHLIGHT {
		/** no highlight applied */
		public static int NONE = 0;
		/** the circle is to be lit. This adjusts location. */
		public static int RING = 1;
		/** the shoulder line is to be lit. This adjusts height / circle radius. */
		public static int LINE = 2;
		/** the left endpoint of the shoulder line is to be lit. This adjusts width. */
		public static int LEFT_POINT = 3;
		/** the right endpoint of the shoulder line is to be lit. This adjusts width. */
		public static int RIGHT_POINT = 4;
		/** the north handle is to be lit. This adjusts orientation. */
		public static int NORTH_HANDLE = 5;
	}

	private OmegaHeadModel localCopy;
	private double x;
	private double y;
	private double lineLength;
	private double diameter;
	private double angle; // in radians
	
	// api access
	private Attribute attr;
	private ViperViewMediator mediator;

	
	// Displayed shape
	private PPath shoulderLine = new PPath();
	private PPath headCircle = new PPath();
	private PPath northPt = new PPath(); 
	private Line2D.Double line;
	private Ellipse2D.Double circle;

	// UI
	private Point2D northHandlePt;
	private int cornerRadius = 5;
	private int selectionThreshold = 5;
	
	// For highlighting
	private PPath highlight = new PPath();
	private int currentHightlight = HIGHLIGHT.NONE;

	public OmegaNode() {
		this(null);
	}
	
	/**
	 *  
	 */
	public OmegaNode(ViperViewMediator mediator) {
		this.mediator = mediator;

		// Set handle as obox child
		addChild(shoulderLine);
		addChild(headCircle);
		addChild(northPt);
		addChild(highlight);

		resetStyle();
	}

	/**
	 * 
	 */
	protected void resetStyle() {
		shoulderLine.setStroke(getDisplayProperties().getStroke());
		headCircle.setStroke(getDisplayProperties().getStroke());
		shoulderLine.setStrokePaint(getDisplayProperties().getStrokePaint());
		headCircle.setStrokePaint(getDisplayProperties().getStrokePaint());
		
		highlight.setStroke(getHighlightDisplayProperties().getStroke());
		highlight.setStrokePaint(getHighlightDisplayProperties().getStrokePaint());
		
		northPt.setStroke(getHandleDisplayProperties().getStroke());
		northPt.setStrokePaint(getHandleDisplayProperties().getStrokePaint());
	}

	/**
	 * Can't use getWidth() since PNode already defines it
	 * 
	 * @return the width of obox
	 */
	public double getOmegaHeight() {
		return diameter;
	}

	public void setShapeAngleInRadians(double angleIn) {
		angle = angleIn;
		updateOmega();
	}

	public double getAngleInRadians() {
		return angle;
	}

	public void setWidthAndHeight(double widthIn, double heightIn) {
		this.lineLength = widthIn;
		this.diameter = heightIn;
		updateOmega();
	}

	public void updateOmega() {
		updateOmegaPart();
		updateHandles();
	}

	private AffineTransform getTransformForRotate() {
		AffineTransform transFrom = AffineTransform
				.getTranslateInstance(-x, -y);
		AffineTransform rotate = AffineTransform.getRotateInstance(-angle);
		AffineTransform transBack = AffineTransform.getTranslateInstance(x, y);
		rotate.concatenate(transFrom);
		transBack.concatenate(rotate);
		return transBack;
	}

	private void updateOmegaPart() {
		double yDistance = diameter/2;
		double xDistance = lineLength/2;
		double cosA = Math.cos(angle);
		double sinA = Math.sin(angle);
		double xdelY = yDistance * sinA; double ydelY = yDistance * cosA;
		double xdelX = xDistance * cosA; double ydelX = xDistance * sinA;
		double x1 = x + xdelY - xdelX;
		double x2 = x + xdelY + xdelX;
		double y1 = y + ydelY + ydelX;
		double y2 = y + ydelY - ydelX;
		line = new Line2D.Double(x1, y1, x2, y2);
		shoulderLine.setPathTo(line);
		circle = new Ellipse2D.Double(x-yDistance, y-yDistance, diameter, diameter);
		headCircle.setPathTo(circle);
		double distance = (diameter/2) + HANDLE_LENGTH;
		northHandlePt = new Point2D.Double(x - distance * sinA, y - distance * cosA);
		northPt.setPathTo(new Ellipse2D.Double(northHandlePt.getX()-.5, northHandlePt.getY()-.5, 1, 1));
	}

	private void updateHandles() {
		int r = 8;
		Point2D c = null;
		switch (currentHightlight) {
		case HIGHLIGHT.RING: 
			highlight.setPathTo(headCircle.getPathReference());
			break;
		case HIGHLIGHT.LINE: 
			highlight.setPathTo(shoulderLine.getPathReference());
			break;
		case HIGHLIGHT.LEFT_POINT: 
			c = getLeftPoint();
			break;
		case HIGHLIGHT.RIGHT_POINT:
			c = getRightPoint();
			break;
		case HIGHLIGHT.NORTH_HANDLE:
			// The north handle is used to adjust orientation
			// It should be HANDLE_LENGTH away from the top of the omega
			c = northHandlePt;
			break;
		case HIGHLIGHT.NONE:
		default:
			highlight.setPathTo(EMPTY_SHAPE);
		}
		if (c != null) {
			double x = c.getX(); double y = c.getY();
			highlight.setPathTo(new Ellipse2D.Double(x-r, y-r, 2*r, 2*r));
		}
	}

	public Object getUpdatedAttribute() {
		int degrees = (int) Math.toDegrees(angle);
		if (degrees < 0) {
			degrees += 360 * (1 + -degrees/360);
		}
		degrees = degrees % 360;
		if (mediator != null) {
			Instant now = mediator.getMajorMoment();
			// Get the oriented box corresponding to current frame
			OmegaHeadModel origBox = (OmegaHeadModel) attr.getAttrValueAtInstant(now);

			// To make sure tiny changes don't affect result
			if (origBox != null) {
				int setDegrees = origBox.getOrientation();
				if (Math.abs(degrees - setDegrees) < 1.2)
					degrees = origBox.getOrientation();
			}
		}

		OmegaHeadModel newBox = new OmegaHeadModel(new Pnt((int) x,
				(int) y), (int) lineLength, (int) diameter,
				degrees);
		return newBox;
	}

	public void setCornerRadius(int val) {
		cornerRadius = val;
	}

	int getCornerRadius() {
		return cornerRadius;
	}

	public void setAttribute(Attribute attr) {
		this.attr = attr;
		Instant now = mediator.getMajorMoment();
		localCopy = (OmegaHeadModel) attr.getAttrValueAtInstant(now);
		// Extract information about circle for local use
		if (localCopy != null) {
			this.angle = Math.toRadians(localCopy.getOrientation());
			this.diameter = localCopy.getDiameter();
			this.lineLength = localCopy.getLineLength();
			this.x = localCopy.getCentroid().getX().intValue();
			this.y = localCopy.getCentroid().getY().intValue();
			updateOmega();
		}
	}
	public Attribute getAttribute() {
		return attr;
	}
	
	/**
	 * 
	 * @return Returns the currentHightlight.
	 */
	public int getCurrentHightlight() {
		return currentHightlight;
	}
	/**
	 * @param currentHightlight The currentHightlight to set.
	 */
	public void setCurrentHightlight(int currentHightlight) {
		this.currentHightlight = currentHightlight;
		updateHandles();
	}

	public Point2D getLeftPoint() {
		return line.getP1();
	}
	public Point2D getRightPoint() {
		return line.getP2();
	}
	
	/**
	 * Gets the region the point is hovering over.
	 * @param selectPt
	 * @return
	 */
	public int getHoverRegion(Point2D selectPt) {
		if (selectPt.distance(getLeftPoint()) < selectionThreshold) {
			return HIGHLIGHT.LEFT_POINT;
		} else if (selectPt.distance(getRightPoint()) < selectionThreshold) {
			return HIGHLIGHT.RIGHT_POINT;
		} else if (circle.contains(selectPt)) {
			return HIGHLIGHT.RING;
		} else if (line.ptSegDist(selectPt) < selectionThreshold) {
			return HIGHLIGHT.LINE;
		} else if (selectPt.distance(northHandlePt) < selectionThreshold) {
			return HIGHLIGHT.NORTH_HANDLE;
		} else if (selectPt.distance(x, y) - diameter/2 < selectionThreshold) {
			return HIGHLIGHT.RING;
		}
		return HIGHLIGHT.NONE;
	}
	
	/**
	 * Calculates the distance from the point to the shape.
	 * If the point is within the shape, returns zero.
	 * @param pt
	 * @return
	 */
	public double ptShapeDist(Point2D pt) {
		double d = Math.min(line.ptSegDist(pt), pt.distance(x, y) - (diameter/2));
		d = Math.min(d, pt.distance(northHandlePt));
		return Math.max(0, d);
	}
	
	public int getSelectionThreshold() {
		return selectionThreshold;
	}
	public void setSelectionThreshold(int selectionThreshold) {
		this.selectionThreshold = selectionThreshold;
	}

	/**
	 * @return
	 */
	public Line2D getLine() {
		return line;
	}

	/**
	 * @param d
	 */
	public void setShapeX(double x) {
		this.x = x;
		updateOmega();
	}
	/**
	 * @param d
	 */
	public void setShapeY(double y) {
		this.y = y;
		updateOmega();
	}
	/**
	 * @param d
	 */
	public double getShapeX() {
		return x;
	}
	/**
	 * @param d
	 */
	public double getShapeY() {
		return y;
	}

	/**
	 * @return
	 */
	public double getShapeDiameter() {
		return this.diameter;
	}

	/**
	 * @param d
	 */
	public void setShapeDiameter(double d) {
		this.diameter = d;
		updateOmega();
	}

	/**
	 * @return
	 */
	public Point2D getNorthHandlePoint() {
		return this.northHandlePt;
	}

	/**
	 * @return
	 */
	public double getShapeLineLength() {
		return this.lineLength;
	}

	/**
	 * @param w
	 */
	public void setShapeLineLength(double w) {
		this.lineLength = w;
		updateOmega();
	}
	
	/**
	 * @return Returns the mediator.
	 */
	public ViperViewMediator getMediator() {
		return mediator;
	}
	
	/**
	 * The mediator is required for 'setAttribute' method, and for
	 * making sure that the rotation is only changed when there is 
	 * a large enough change.
	 * @param mediator The mediator to set.
	 */
	public void setMediator(ViperViewMediator mediator) {
		this.mediator = mediator;
	}
}